# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging

from availsim4core.src.context.context import Context
import itertools
import numpy as np

from typing import List, Set
from availsim4core.src.context.system.children_logic.tolerated_fault import ToleratedFault

from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.oo import Oo
from availsim4core.src.simulation.importance_splitting.simple_weights import SimpleWeights


class ChildrenLogicCriticalFailurePathsError(Exception):
    pass


class SpecificComponentNotFoundCriticalFailurePathsError(Exception):
    pass


class CriticalFailurePaths:
    """
    Class used to gather information about minimal critical failure paths. Creation of an object of this class
    triggers analysis of the component tree to identify combinations of components leading to critical failures
    of the overall system. The overall system is identified by a system root component selected by users in the
    simulation input sheet (in anticipation of systems which root is not identical with the AvailSim4 component
    tree).
    """

    def __init__(self, context: Context, system_root_component_name: str):
        self.context = context
        basic_components = [component for component in context.root_component.to_set() if isinstance(component, Basic)]
        global_to_basic_id_dict = {}
        basic_to_global_id_dict = {}
        for idx, component in enumerate(basic_components):
            global_to_basic_id_dict[component.global_id] = idx
            basic_to_global_id_dict[idx] = component.global_id
        self.n_basic_components = len(basic_components)
        self.global_to_basic_id_dict = global_to_basic_id_dict
        self.basic_to_global_id_dict = basic_to_global_id_dict

        self.cfp_analysis_starting_node = self.find_system_root_component(
            system_root_component_name)
        self.critical_failure_paths = self.generate_critical_failure_paths(
            self.cfp_analysis_starting_node)

        self.longest_cfp_len = max(
            [len(failure_path) for failure_path in self.critical_failure_paths])
        self.cfp_matrix = self.fetch_cfp_matrix(self.critical_failure_paths)
        self.weights = SimpleWeights
        self.component_weights = self.calculate_component_failure_weight_vector(
            context.root_component.to_set())



    def find_system_root_component(self, system_root_component_name: str) -> Component:
        """
        This function returns a component object from the tree which has the same name as
        the system_root_component_name parameter decided by the user in the simulation input file. If the component
        is not found, an exception is logged and a SpecificComponentNotFoundCriticalFailurePathsError is raised.
        Both contain a message which states the missing name recovered during execution.
        """
        for component in self.context.root_component.to_set():
            if component.name == system_root_component_name:
                return component
        message_exception = f"The component name {system_root_component_name} specified for the CRITICAL_FAILURE_PATHS " \
                            f"diagnostic has not be found"
        logging.exception(message_exception)
        raise SpecificComponentNotFoundCriticalFailurePathsError(
            message_exception)

    @staticmethod
    def getXooYfailureCombinations(
            n: int,
            failure_paths_of_children_grouped: List[List[Component]],
            failure_paths: List[List[Component]]):
        # n is the number of components that have to work. n = 1 is essentially OR logic.
        # We get children_n groups of components causing critical failures of each of the children.
        # children_cfp is children_n lists of critical failure paths for each children
        # cfp is the returned result

        # XooY is referring to the way failure logic is defined. X stands for the number of components
        # that are required to work, and Y for the number of all children components
        x = int(n)
        y = len(failure_paths_of_children_grouped)

        # creating all possible combinations of choosing X groups of failure paths out of Y children groups
        failure_groups_combinations = [list(x) for x in list(
            itertools.combinations(failure_paths_of_children_grouped, y-x+1))]

        for failure_group_combination in failure_groups_combinations:
            # identifying the value of a largest  path created in all cobinations
            max_path_len = max([len(failure_path)
                               for failure_path in failure_group_combination])
            indices = list(itertools.product(
                range(0, max_path_len), repeat=len(failure_group_combination)))

            for index_set in indices:
                new_failure_paths = []
                for idx, i in enumerate(index_set):
                    if i < len(failure_group_combination[idx]):
                        new_failure_paths.extend(failure_group_combination[idx][i])
                    else:
                        new_failure_paths = []
                        break
                if new_failure_paths:
                    failure_paths.append(new_failure_paths)

    def generate_critical_failure_paths_recursive_call(self, component: Component) -> List[List[Component]]:
        """
        Recursive function parsing the entire tree
        :param component: component recursively explored
        :return cfp: the list of failure paths for the component inputed
        """

        # recursive part
        children_cfp = []
        for child in component.get_children():
            children_cfp.append(
                self.generate_critical_failure_paths_recursive_call(child))

        # handling the failure paths at the current level, according to the children logic
        cfp = []
        if not component.get_children():
            cfp.append([component])
        else:
            if isinstance(component.children_logic, And):
                for children_group in children_cfp:
                    cfp.extend(children_group)
            elif isinstance(component.children_logic, ToleratedFault):
                self.getXooYfailureCombinations(len(component.get_children()) - component.children_logic.fault_tolerance,
                                                children_cfp, cfp)
            elif isinstance(component.children_logic, Oo):
                # the extend of cfp is done within the function getXooYfailureCombinations
                self.getXooYfailureCombinations(
                    component.children_logic.minimum_number_of_required_component, children_cfp, cfp)
            else:
                message_exception = f"The children logic {component.children_logic} of component {component} is not " \
                                    f"implemented in the function generate_critical_failure_paths_recursive_call"
                logging.exception(message_exception)
                raise ChildrenLogicCriticalFailurePathsError(message_exception)

        return cfp

    def search_for_minimal_paths(self, input_list_of_lists_of_components: List[List[Component]]) -> List[List[Component]]:
        """
        Function parsing a list of list of elements and spotting if any list is included in an other list. The goal is
        to find "minimal failure critical path" so if a path A is included in B it means that B is not a minimal paths.
        :param cfp: list of list of components representing critical paths
        :return cfp: cleaned list of list of components
        """

        # transforming the list of lists into a list of sets to avoid duplicates within one list
        # and better check the inclusion
        list_of_sets_of_components = []
        for path in input_list_of_lists_of_components:
            list_of_sets_of_components.append(set(path))

        # nested loops used to compare two elements in the list
        output_list_of_lists_of_components = []
        for ida, A_path in enumerate(list_of_sets_of_components):
            A_path_is_minimal = True
            for idb, B_path in enumerate(list_of_sets_of_components):
                if ida != idb:  # if ida == idb then A == B then A include in B and B included in A
                    if B_path.issubset(A_path):
                        # A_path is not a minimal path because B_path is included in A_path
                        A_path_is_minimal = False
                        break

            if A_path_is_minimal:
                output_list_of_lists_of_components.append(list(A_path))

        return output_list_of_lists_of_components

    def generate_critical_failure_paths(self, component: Component) -> List[List[Component]]:
        """
        This function aggregates calls to two other functions: one which is responsible for identifying all
        critical failure paths in the system (recursively) and one which ensures that the identified
        combinations are not repeated (also within other combinations). The taken argument points
        to the start of the tree and the function returns a list of minimal critical failure paths.
        """
        critical_failure_paths = self.generate_critical_failure_paths_recursive_call(component)
        critical_failure_paths = self.search_for_minimal_paths(critical_failure_paths)

        return critical_failure_paths

    def fetch_cfp_matrix(self, results) -> np.ndarray:
        """
        This function creates a matrix describing critical failure paths as consecutive vectors of
        true/false values. Such a representation is going to be useful for fast criticallity lookup.
        """
        cfp_matrix = np.zeros([len(results), self.n_basic_components])
        for idx, result in enumerate(results):
            for component in result:
                cfp_matrix[idx,
                           self.global_to_basic_id_dict[component.global_id]] = 1
        return cfp_matrix

    def calculate_component_failure_weight_vector(self, components_set: Set[Component]) -> np.ndarray:
        """
        This function prepares an integer vector representation of weights for components in the system.
        Each individual basic component is assigned a value according to the formula in this function and
        then stored in a vector on a position assigned to the component. Such a representation is applied
        to improve the speed of responses for fast criticallity lookups.
        """
        weight_vector = np.zeros(self.n_basic_components)
        for component in components_set:
            if isinstance(component, Basic):
                weight = self.weights.calculate_component_failure_weight(
                    component)
                weight_vector[self.global_to_basic_id_dict[component.global_id]] = weight
        return weight_vector

    def calculate_failures_atm_vector(self, components_set: Set[Component]) -> np.ndarray:
        """
        The basic components provided in a set as the only argument of this function are then checked for blind
        failures. Their statuses in this aspect decide on the value in a true/false vector describing the current
        state of the system. Positions of each component in that vector are fixed.
        """
        failure_states_atm_vector = np.zeros(self.n_basic_components)
        for component in components_set:
            if self.weights.is_cfp_component(component):
                failure_states_atm_vector[self.global_to_basic_id_dict[component.global_id]] = 1
        return failure_states_atm_vector

    def calculate_criticallity_atm(self, failure_state_atm: np.ndarray) -> np.ndarray:
        """
        This function is a shorthand for a dot operation between argument failure_state_atm which is a matrix
        expected to contain failure states at the moment and component weights vector.
        """
        return np.dot(failure_state_atm, self.component_weights)

    def calculate_distance_to_critical_failure(self, failure_states_atm: np.ndarray) -> float:
        """
        This function is expected to be extended in the future. For now, its role is limited to counting a number
        of blind failures in the system (by summing true/false vector). The only extension in the current state is
        the condition on the number of failures: whenever there is more failures in the system than there is in the
        longest minimal critical failure path, the result is 0.  This makes sure that whenever cfp_threshold is set
        to a non-zero natural number (which triggers the ISp and thus CFP approaches), the branching will not occur.
        """
        result = self.weights.calculate_distance_to_critical_failure(failure_states_atm)
        if result >= self.longest_cfp_len:
            return 0
        return result



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas
import os


class Result:
    """
    class used to defined a results to compare with the results in the results sheet of the output file.
    an expected result is defined by its expected value plus where it is standing in the sheet.
    also if a sensitivity analysis is performed, an id of the output file has to be defined.
    """

    def __init__(self,
                 expected_result,
                 component_name,
                 phase_name,
                 status_name,
                 column_name,
                 file_number=0,
                 tolerance=-1):
        self.expected_result = expected_result
        self.component_name = component_name
        self.phase_name = phase_name
        self.status_name = status_name
        self.column_name = column_name
        self.file_number = file_number
        self.tolerance = tolerance

    def __str__(self):
        return (
            f"file {self.file_number}; {self.component_name}, {self.phase_name}, {self.status_name}, {self.column_name} = "
            f"= {self.expected_result} up to {self.tolerance}th digit")

    def extract_result(self,
                       result_simulation_file_list):

        for result_simulation_file in result_simulation_file_list:
            # result_simulation_file_number = int(result_simulation_file.split("/")[-1].split("_")[2])
            result_simulation_file_number = int(result_simulation_file.split(os.sep)[-1].split("_")[2])

            if self.file_number == result_simulation_file_number:
                data = pandas.read_excel(result_simulation_file,
                                         sheet_name="RESULTS",
                                         engine='openpyxl')

                if self.component_name == "*":
                    data_loc = data.copy()
                else:
                    data_loc = data.loc[(data.component == self.component_name)]

                if not self.phase_name == "*":
                    data_loc = data_loc.loc[(data.phase == self.phase_name)]

                if not self.status_name == "*":
                    data_loc = data_loc.loc[(data.status == self.status_name)]

                result = float(data_loc[self.column_name].sum())

        return result


class RCAResult:
    """
    class defining a result to compare with the results in the RCA results sheet of the output file.
    """

    def __init__(self,
                 expected_result,
                 component_name,
                 phase_name,
                 status_name,
                 file_number=0,
                 tolerance=-1):
        self.expected_result = expected_result
        self.component_name = component_name
        self.phase_name = phase_name
        self.status_name = status_name
        self.file_number = file_number
        self.tolerance = tolerance

    def __str__(self):
        return (
            f"file {self.file_number}; {self.component_name}, {self.phase_name}, {self.status_name} = "
            f"= {self.expected_result} up to {self.tolerance}th digit")

    def extract_result(self,
                       result_simulation_file_list):

        for result_simulation_file in result_simulation_file_list:
            result_simulation_file_number = int(result_simulation_file.split("/")[-1].split("_")[2])

            if self.file_number == result_simulation_file_number:
                data = pandas.read_excel(result_simulation_file,
                                         sheet_name="RESULTS_ROOT_CAUSE_ANALYSIS",
                                         engine='openpyxl')

                if len(data)>0:

                    if self.component_name == "*":
                        data_loc = data.copy()
                    else:
                        data_loc = data.loc[(data.rca_component_trigger == self.component_name)]

                    if not self.phase_name == "*":
                       data_loc = data_loc.loc[(data.rca_phase_trigger == self.phase_name)]

                    if not self.status_name == "*":
                        data_loc = data_loc.loc[(data.rca_status_trigger == self.status_name)]

                    result = len(data_loc)

                else:

                    result = 0

        return result


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from enum import Enum
from typing import List, Set, Tuple

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent
from availsim4core.src.discrete_event_simulation.event.event import Event
from availsim4core.src.timeline.record import Record


class BEventPriority(Enum):
    """
    class used to easily order the B Events when they are occuring exactly at the same timestamp
    """

    BLIND_FAILURE_EVENT = -2
    DETECTABLE_FAILURE_EVENT = -1

    JUMP_PHASE_EVENT = 0
    NEXT_PHASE_IF_SPECIFIC_FAILURE_EVENT = 1
    NEXT_PHASE_IF_FAILURE_EVENT = 2
    NEXT_PHASE_EVENT = 3

    END_HOLDING_EVENT = 4

    END_REPAIRING_EVENT = 5
    MRU_END_REPAIRING_EVENT = 6

    START_REPAIRING_EVENT = 7
    MRU_START_REPAIRING_EVENT = 8

    END_INSPECTION_EVENT = 9
    START_INSPECTION_EVENT = 10

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class BEvent(Event):
    """
    B  events  =  Bound  events,  also  called  Timed  events,  Unconditioned  events.
    B event only occurs on the Basic components.
    Those events are planned, do not need any conditions to be fulfilled in order to happen but just occur due to a delay being expired.
    In our case, it would be a failure occurring or a repair finishing (and its propagation in the system), a periodic inspection starting, the end of a phase.
    """
    __slots__ = 'absolute_occurrence_time', 'priority'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 priority: BEventPriority):
        super().__init__(context)
        self.absolute_occurrence_time = absolute_occurrence_time
        self.priority = priority

    def __eq__(self, other):
        return self.absolute_occurrence_time == other.absolute_occurrence_time and \
               self.priority == other.priority

    def __lt__(self, other):
        return self.absolute_occurrence_time < other.absolute_occurrence_time or \
               (self.absolute_occurrence_time == other.absolute_occurrence_time and self.priority < other.priority)

    def is_context_valid(self):
        return True

    def postpone(self, duration):
        """
        This function is only implemented for blind failure event and detectable failure event.
        :param duration: ttf := ttf + duration
        :return: event
        """
        return self

    def execute(self) -> List[Record]:
        """
        update the status of a component by calling update_status
        the new status depend on the type of B event
        """
        return []

    def generate_c_event(self, **kwargs) -> Set[CEvent]:
        """
        Returns the set of CEvents generated by this particular b_event.
        """
        return set()

    def update_b_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """
        event_deleted: Set = set()
        return event_set, event_deleted

    def update_c_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """
        event_deleted: Set = set()
        return event_set, event_deleted

    @staticmethod
    def _b_events_to_be_cleaned() -> List[Event]:
        """
        Type list of Events that will be affected by this particular event.
        """
        return []

    @staticmethod
    def _c_events_to_be_cleaned() -> List[Event]:
        """
        Type list of Events that will be affected by this particular event.
        """
        return []


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class SystemModifierCombinationStrategy:

    @classmethod
    def execute(cls, exploration_strategy_list):
        pass

    @staticmethod
    def filter_exploration(exploration_strategy_list,
                           exploration_class):
        """
        Given an {ExplorationStrategy} list it will filter and return only the elements of a given exploration class.
        :param exploration_strategy_list list of {ExplorationStrategy} to filter
        :param exploration_class the {ExplorationStrategy} class to filter out.
        """
        return [
            exploration_strategy
            for exploration_strategy in exploration_strategy_list
            if isinstance(exploration_strategy, exploration_class)
        ]



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_MRU(unittest.TestCase):
    """
    Ensemble of tests used to test the Minimal Replaceable Unit feature
    """

    expected_result_scenario_1 = [Result(3, "B1_0_3", "NONE", "FAILED", "_MEAN_OCCURRENCES")]
    scenario_1_no_MRU = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/no_active_mru_system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    expected_result_scenario_2 = [Result(2, "B1_0_3", "NONE", "FAILED", "_MEAN_OCCURRENCES")]
    scenario_2_basic_trigger_itself = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/basic_trigger_itself_system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    expected_result_scenario_3 = [Result(3, "A2_0_11", "NONE", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(3, "B1_0_3", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "B2_0_4", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "B3_0_6", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "B3_1_7", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES")]
    scenario_3_basic_trigger = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/basic_trigger_system.xlsx",
         None,
         1,
         expected_result_scenario_3]

    expected_result_scenario_4 = [Result(3, "A2_0_11", "NONE", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(3, "B1_0_3", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "B2_0_4", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "B3_0_6", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "B3_1_7", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES")]
    scenario_4_compound_trigger = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/compound_trigger_system.xlsx",
         None,
         1,
         expected_result_scenario_4]

    # multiple MRU in the system
    # one component trigger its own replacement plus other replacements
    expected_result_scenario_5_v1 = [Result(2, "D_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(1, "D_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "D_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(2, "C1_0_2", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(1, "C1_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "C1_0_2", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(2, "C2_0_3", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "C2_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "C2_0_3", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(1, "C3_0_4", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "C3_0_4", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(0, "C3_0_4", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  ]
    scenario_5_v1 = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/multiple_mru_v1_system.xlsx",
         None,
         1,
         expected_result_scenario_5_v1]

    # multiple MRU in the system
    # one component trigger its own replacement plus other replacements, BLIND_FAILURES used
    expected_result_scenario_5_v2 = [Result(2, "D_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                     Result(1, "D_0_1", "*", "BLIND_FAILED", "_MEAN_OCCURRENCES"),
                                     Result(0, "D_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                     Result(1, "D_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                     Result(2, "C1_0_2", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                     Result(1, "C1_0_2", "*", "BLIND_FAILED", "_MEAN_OCCURRENCES"),
                                     Result(0, "C1_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                     Result(1, "C1_0_2", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                     Result(2, "C2_0_3", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                     Result(0, "C2_0_3", "*", "BLIND_FAILED", "_MEAN_OCCURRENCES"),
                                     Result(0, "C2_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                     Result(1, "C2_0_3", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                     Result(1, "C3_0_4", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                     Result(0, "C3_0_4", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                     Result(0, "C3_0_4", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  ]
    scenario_5_v2 = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/multiple_mru_v2_system.xlsx",
         None,
         1,
         expected_result_scenario_5_v2]

    # multiple MRU in the system
    # one component trigger several MRU for several components
    expected_result_scenario_5_v3 = [Result(2, "D_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(1, "D_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "D_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(2, "C1_0_2", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(1, "C1_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "C1_0_2", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(2, "C2_0_3", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "C2_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "C2_0_3", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(2, "C3_0_4", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "C3_0_4", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "C3_0_4", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(1, "C4_0_5", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "C4_0_5", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(0, "C4_0_5", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  ]
    scenario_5_v3 = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/multiple_mru_v3_system.xlsx",
         None,
         1,
         expected_result_scenario_5_v3]

    # multiple MRU in the system
    # idem than v3, only some permutations in the input are performed but they should not have any impacts
    scenario_5_v4 = \
        ["./availsim4core/test/E2E/input/mru/simulation.xlsx",
         "./availsim4core/test/E2E/input/mru/multiple_mru_v4_system.xlsx",
         None,
         1,
         expected_result_scenario_5_v3]

    param_scenario = [
        scenario_1_no_MRU,
        scenario_2_basic_trigger_itself,
        scenario_3_basic_trigger,
        scenario_4_compound_trigger,
        scenario_5_v1,
        scenario_5_v2,
        scenario_5_v3,
        scenario_5_v4,
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_result_list in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_result_list)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    sensitivity_file,
                    expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/mru/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)
            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from abc import abstractmethod
from typing import List, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from availsim4core.src.context.context import Context
    from availsim4core.src.context.system.component_tree.compound import Compound

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.timeline.record_component import RecordComponent


class ComponentType:
    BASIC = "BASIC"
    COMPOUND = "COMPOUND"


class Component:
    """
    Defines a component of the MUTABLE Tree system structure. It can be specified as Compound or Basic.
    The tree is mutable (on the Discrete event simulation level, status and statistics_des)
    because it needs to handle million of iteration when manipulating the status and evaluating the statistics.
    """
    __slots__ = 'global_id', 'name', 'local_id', '_parents', 'status', 'list_of_mru_trigger'

    def __init__(self,
                 global_id: int,
                 name: str,
                 local_id: int,
                 parents: List[Compound],
                 list_of_mru_trigger: List[MinimalReplaceableUnit]):
        self.global_id = global_id
        self.local_id = local_id
        self.name = name
        self._parents: List[Compound] = parents
        self.status = Status.RUNNING
        self.list_of_mru_trigger = list_of_mru_trigger

    def __hash__(self):
        return hash(self.global_id)

    def __eq__(self, other):
        return self.global_id == other.global_id

    def __lt__(self, other):
        return self.global_id < other.global_id

    def get_parents(self):
        return self._parents

    def add_parent(self, parent):
        self._parents.append(parent)

    def get_children(self):
        return []

    def remove(self, component) -> None:
        pass

    def is_shared(self):
        return len(self._parents) > 1

    def _propagate_status(self,
                          timestamp: float,
                          phase: Phase,
                          description: str,
                          context: Context):
        """
        Propagate the status to the parents Components.
        :param timestamp: timestamps of the new status
        :param phase: phase in which the new status changed
        :param description: description of the status.
        :param context: the context of the simulation (structure containing almost everything)
        """
        record_list = []
        for compound_parent in self._parents:
            new_record_list = compound_parent.update_status(timestamp, phase, description, context)
            record_list.extend(new_record_list)
        return record_list

    def _update_status(self,
                       new_status: Status,
                       timestamp: float,
                       phase: Phase,
                       description: str,
                       context: Context):
        """
        Update the status of the current Component, if the status changes it is recorded in the statistics timeline
        and propagated to the component parents.
        :param new_status: New status of the Component
        :param timestamp: timestamp of the status
        :param description: description of the new status.
        """
        record_list = []
        if self.status != new_status:
            self.status = new_status
            record = RecordComponent(self, new_status, timestamp, phase, description)
            record_list.append(record)
            new_record_list = self._propagate_status(timestamp, phase, description, context)
            record_list.extend(new_record_list)
        return record_list

    @abstractmethod
    def to_set(self) -> Set:
        """
        :return: The flat Pre-order (NLR) representation of the tree.
        ::see https://en.wikipedia.org/wiki/Tree_traversal
        """
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterListing(ResultExporter):

    @staticmethod
    def component_listing_export(root_component: Component):
        """
        Function exploring the system in order to build a list of components in the system.
        """

        ret = ""

        for component in root_component.to_set():

            if component.get_children():
                # if that component has children (then it is a Compound type), apply the same function to the children
                ret += f"{component.name};{component.local_id};{component.global_id}; compound;" \
                       f" -; -; -; {[x.name for x in component.list_of_mru_trigger]}\n"
            else:
                # else, the current component is a basic and has failure mode(s), those a printed
                ret += f"{component.name};{component.local_id};{component.global_id}; basic;" \
                       f"{component.failure_mode.name}; " \
                       f"{component.failure_mode.failure.type_of_failure}; " \
                       f"{[x.name for x in component.list_of_mru_group]}; " \
                       f"{[x.name for x in component.list_of_mru_trigger]}\n"

        return ret

    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        tree_string = cls.component_listing_export(context.root_component)

        tree_dataframe = pandas.DataFrame([x.split(';') for x in tree_string.split('\n')],columns=[
            "COMPONENT_NAME", "local_id", "global_id", "COMPONENT_TYPE",
            "FAILURE_MODE_NAME", "TYPE_OF_FAILURE",
            "IN_MRU", "TRIGGER_MRU"
        ])

        tree_dataframe.to_excel(writer, sheet_name='RESULTS_COMPONENT_LISTING', index=False)

        writer.save()


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.phase.phase import Phase


class Record:
    """
    Immutable
    """
    __slots__ = 'timestamp', 'phase', 'description'

    def __init__(self,
                 timestamp: float,
                 phase: Phase,
                 description: str):
        self.timestamp = timestamp
        self.phase = phase
        self.description = description

    def __str__(self):
        pass

    def __repr__(self):
        return f"timestamp: {self.timestamp}, description: {self.description}"

    def __eq__(self, other):
        return self.timestamp == other.timestamp \
               and self.phase == other.phase \
               and self.description == other.description

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def get_result_record_entry(self):
        pass

    def check_if_in_rca_triggers(self, rca_triggers, current_phase_name):
        return False



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Component

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.results.result_record_entry_component import ResultRecordEntryComponent
from availsim4core.src.timeline.record import Record
from availsim4core.src.context.system.rca_trigger import RootCauseAnalysisTrigger


class RecordComponent(Record):
    """
    Class of record dedicated to components
    A record is an object used to keep trace of the state of the system
    """
    __slots__ = 'component', 'status'

    def __init__(self,
                 component: Component,
                 status: Status,
                 timestamp: float,
                 phase: Phase,
                 description: str):
        super().__init__(timestamp, phase, description)
        self.component = component
        self.status = status

    def __eq__(self, other):
        if not isinstance(other, RecordComponent):
            return NotImplemented
        return super().__eq__(other) \
               and self.component == other.component \
               and self.status == other.status

    def __repr__(self):
        return super().__repr__() \
               + f" component: {self.component.name} -> id: {self.component.global_id}" \
                 f" status: {self.status} , phase: {self.phase}"

    def __str__(self):
        return f"RecordComponent at {self.timestamp}: " \
               f"{self.component.name}_{self.component.local_id}_{self.component.global_id} " \
               f"with status {self.status} on phase {self.phase.name}__{self.status}"

    def get_result_record_entry(self):
        return ResultRecordEntryComponent(self.component,
                                          self.status,
                                          self.phase,
                                          self.description)

    def check_if_in_rca_triggers(self, rca_triggers, current_phase_name):
        return RootCauseAnalysisTrigger(self.component.name, str(self.component.status), current_phase_name) in rca_triggers


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List


class RootCauseAnalysisTrigger:

    def __init__(self, component_name: str, component_status: str, phase: str):
        self.component_name = component_name
        self.component_status = component_status
        self.phase = phase

    def __str__(self):
        return f"RCA trigger on {self.component_name}, status {self.component_status} in phase {self.phase}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.component_name, self.component_status, self.phase))

    def __eq__(self, other):
        if not isinstance(other, RootCauseAnalysisTrigger):
            return NotImplemented
        return self.component_name == other.component_name and \
               self.component_status == other.component_status and \
               self.phase == other.phase


class RootCauseAnalysisTriggerFactory:
    
    @staticmethod
    def build(triggering_component_name: str,
              triggering_component_statuses: List[str], 
              triggering_phases: List[str]) -> List[RootCauseAnalysisTrigger]:
        return [RootCauseAnalysisTrigger(triggering_component_name, status, phase) 
                for status in triggering_component_statuses 
                for phase in triggering_phases]



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.failure import FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.blind_failure_event import \
    BlindFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
import logging


class FailureEventFactoryError(Exception):
    pass


class FailureEventFactory:
    """
    Utility class to build a 'FailureEvent'.
    """

    @staticmethod
    def build(absolute_time_to_failure,
              context: Context,
              basic: Basic,
              failure_mode: FailureMode):
        """
        Given an absolute_time and a type of failure, it returns the corresponding FailureEvent at the correct timing.
        """

        if failure_mode.failure.type_of_failure == FailureType.DETECTABLE:
            return DetectableFailureEvent(absolute_time_to_failure,
                                          context,
                                          basic,
                                          failure_mode)

        elif failure_mode.failure.type_of_failure == FailureType.BLIND:
            return BlindFailureEvent(absolute_time_to_failure,
                                     context,
                                     basic,
                                     failure_mode)

        else:
            message_exception = f"Wrong type of failure {failure_mode.failure.type_of_failure} in the sheet failureModes"
            logging.exception(message_exception)
            raise FailureEventFactoryError(message_exception)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from typing import Set

import numpy

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.discrete_event_simulation.b_event_generator import BEventGenerator
from availsim4core.src.discrete_event_simulation.c_event_generator import CEventGenerator
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.failure_event import FailureEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent
from availsim4core.src.timeline.record_generator import RecordGenerator
from availsim4core.src.statistics.critical_failure_paths import CriticalFailurePaths


class DiscreteEventSimulation:
    """
    Implementation of the Discrete event simulation following the documentation :
    https://en.wikipedia.org/wiki/Discrete-event_simulation
    """
    __slots__ = 'seed', 'duration_of_simulation', 'context', 'execution_metrics', 'absolute_simulation_time', \
                'root_component_set', 'cfp_threshold', 'critical_failure_paths'

    def __init__(self,
                 seed: int,
                 duration_of_simulation: float,
                 context: Context,
                 cfp: CriticalFailurePaths = None,
                 cfp_threshold: float = 0.0,
                 time: float = 0):

        self.seed = seed
        self.duration_of_simulation = duration_of_simulation
        self.context = context

        self.execution_metrics = {
            "number_of_time_iterations": 0,
            "number_of_b_events_executed": 0,
            "number_of_c_events_executed": 0,
            "number_of_b_events_removed": 0,
            "number_of_c_events_removed": 0,
            "length_of_timeline_record":0,
        }

        self.absolute_simulation_time = time

        self.critical_failure_paths = cfp
        self.cfp_threshold = cfp_threshold

        self.root_component_set = self.context.root_component.to_set()

    def regenerate_events(self, b_events, absolute_simulation_time):
        """
        Whenever a timeline is restarted, exsting b_events are updated to create new timelines. Each event has a corresponding
        attribute defining its occurrence time (`absolute_occurrence_time`), which decides if and when the event takes place.
        This method iterated through all events provided in the argument and re-draws absolute_occurrence_time using the original
        sample_generation_time (in-simulation time when a given sample was created) in respect to which a new sample is created
        using appropriate failure laws and scale parameters.

        The condition checked for each re-drawn sample is that the event occurrs after the current simulation time (second argument,
        `absolute_simulation_time`). Otherwise, the events could be expected to take place before the event triggering the splitting,
        changing the timeline which is to be replicated.
        """

        new_b_events = set()
        for b_event in b_events:
            if isinstance(b_event, FailureEvent): 
                proposed_random_adjustment = b_event.failure_mode.failure_law.get_quantile_value(numpy.random.uniform())
                b_event.absolute_occurrence_time = b_event.failure_mode.sample_generation_time + proposed_random_adjustment
                while b_event.absolute_occurrence_time < absolute_simulation_time:
                    proposed_random_adjustment = b_event.failure_mode.failure_law.get_quantile_value(numpy.random.uniform())
                    b_event.absolute_occurrence_time = b_event.failure_mode.sample_generation_time + proposed_random_adjustment
            new_b_events.add(b_event)
        return new_b_events

    def __eq__(self, other):
        return self.seed == other.seed and \
               self.duration_of_simulation == other.duration_of_simulation and \
               self.context == other.context and \
               self.execution_metrics == other.execution_metrics and \
               self.absolute_simulation_time == other.absolute_simulation_time and \
               self.b_events_set == other.b_events_set and \
               self.c_events_set == other.c_events_set and \
               self.root_component_set == other.root_component_set

    def run(self):
        """
        This method runs the three phase approach of the discrete simulation algorithm.
        The discrete event simulation runs step by step from event to event until it reaches
        the 'duration of simulation'.
        Note: each event is evaluated at a time. If several events are happening at the exact same time,
        they will be processed one by one.
        """

        numpy.random.seed(int(self.seed))

        if not (self.context.b_events_set or self.context.c_events_set):
            # if b events set or c events set are both empty, then the simulation is starting from the beginning
            # and thus needs to initialize first events in the simulation
            init_record = RecordGenerator.get_initial_records(self.absolute_simulation_time,
                                                            self.context)
            self.context.timeline_record.record_list.extend(init_record)
            self.context.b_events_set = BEventGenerator.generate_first_events(self.absolute_simulation_time,
                                                                  self.context)
            self.context.c_events_set = CEventGenerator.generate_first_events(self.absolute_simulation_time,
                                                                  self.context)
            # first execution of c_events in order to generate some additional b_events
            self._execute_c_events()
        else:
            # when there are some b or c events, the simulation is not starting at the begininng, which means that we
            # are restarting inside another timeline
            self.context.b_events_set = self.regenerate_events(self.context.b_events_set, self.absolute_simulation_time)

        # reached_old_threshold = False

        while True:

            logging.debug(f"Discrete event simulation iteration: {self.execution_metrics['number_of_time_iterations']}")
            logging.debug(f"Discrete event simulation absolute time: {self.absolute_simulation_time}")

            logging.debug(f"BEvents list: {self.context.b_events_set}")
            logging.debug(f"CEvents list: {self.context.c_events_set}")

            if not self.context.b_events_set:
                break

            if(self.cfp_threshold > 0.0):
                # when cfp_threshold is above 0.0, check the ISp thresholds and possibly trigger more branches of the simulation
                # - define failure_states_atm
                failure_states_atm_vector = self.critical_failure_paths.calculate_failures_atm_vector(self.root_component_set)

                # - calculate the failure_criticallity matrix
                failure_criticallity = self.critical_failure_paths.calculate_distance_to_critical_failure(failure_states_atm_vector)

                if(failure_criticallity >= self.cfp_threshold):
                    return self.context, self.absolute_simulation_time

            b_events_valid_context = {b_event
                                      for b_event in self.context.b_events_set
                                      if b_event.is_context_valid()}

            current_b_event = min(b_events_valid_context)

            self.absolute_simulation_time = current_b_event.absolute_occurrence_time

            if self.absolute_simulation_time > self.duration_of_simulation:
                break

            logging.debug(f"Current valid BEvent: {current_b_event}")

            self.execution_metrics['number_of_time_iterations'] += 1

            self._execute_b_events(b_events_valid_context)

            self._execute_c_events()

        self.execution_metrics["length_of_timeline_record"]=len(self.context.timeline_record.record_list)

        return self.context, self.absolute_simulation_time #self.context.timeline_record

    def _execute_b_events(self, b_events_valid_context):
        """
        Manage the all the simultaneous B events.
        For each of the b_events_set, it tests if the b_event is happening at the current time.
        """
        b_events_simultaneous = {b_event
                                 for b_event in b_events_valid_context
                                 if b_event.absolute_occurrence_time == self.absolute_simulation_time}

        logging.debug(f"Simultaneous BEvent list: {b_events_simultaneous}")
        for b_event in sorted(b_events_simultaneous):
            if b_event in self.context.b_events_set:
                self._execute_b_event(b_event)

        self.context.b_events_set = self.context.b_events_set - b_events_simultaneous

    def _execute_b_event(self, b_event: BEvent):
        """
        Given a b_event, this method:
         - update the component status (and propagate the status)
         - update the occurrence time of the for each component of the b_event_list
         - generate a list of c_event related to the current b_event and update the c_event_set.
        :param b_event: The b_event to execute.
        """

        logging.debug(f"Executed BEvent: {b_event}")
        record_list = b_event.execute()
        self.execution_metrics['number_of_b_events_executed'] += 1
        logging.debug(f"Records list: {record_list}")
        self.context.timeline_record.record_list.extend(record_list)

        # Diagnosis - checking the RCA dump triggers
        self.context.rca_manager.trigger_root_cause_analysis_check(record_list,
                                                                self.seed,
                                                                self.context.phase_manager.current_phase.name)

        self.context.b_events_set, b_event_removed_set = b_event.update_b_event_collection(
            self.context.b_events_set,
            b_event._b_events_to_be_cleaned()
        )

        self.execution_metrics["number_of_b_events_removed"]+=len(b_event_removed_set)

        self.context.c_events_set, c_event_removed_set = b_event.update_c_event_collection(
            self.context.c_events_set,
            b_event._c_events_to_be_cleaned()
        )

        self.execution_metrics["number_of_c_events_removed"]+=len(c_event_removed_set)

        new_c_event_set = b_event.generate_c_event(b_event_removed_set=b_event_removed_set)

        logging.debug(f"Generated CEvent list: {new_c_event_set}")

        self.context.c_events_set.update(new_c_event_set)

    def _execute_c_events(self):
        """
        Manage the C events.
        For each of the c_events_list, it tests if the condition is valid and apply it.
        If the result of this valid c_event generates b_event, then the b_event_list is updated and sorted accordingly.
        """
        valid_c_events = {event
                          for event in self.context.c_events_set
                          if event.is_condition_valid()}

        for valid_c_event in sorted(valid_c_events):
            logging.debug(f"Executed CEvent with condition valid: {valid_c_event}")
            new_b_events_set = valid_c_event.generate_b_events(self.absolute_simulation_time)
            self.execution_metrics['number_of_c_events_executed'] += 1
            logging.debug(f"Generated BEvents: {new_b_events_set}")
            self.context.b_events_set.update(new_b_events_set)

        self.context.c_events_set = self.context.c_events_set - valid_c_events



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from datetime import datetime

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.rca.rca_manager import RootCauseAnalysisManager
from availsim4core.src.context.system.component_tree.component_factory import ComponentFactory
from availsim4core.src.exporter.aggregate_exporter import AggregateExporter


class LocalRunner:
    """
    class used to run simulation on the local machine (where the code is called)
    """

    @staticmethod
    def run(output_folder,
            analysis: Analysis):
        start_time = datetime.now()
        logging.info(f"Starting Analysis {analysis.id}")

        component_factory = ComponentFactory(analysis.system_template)
        root_component = component_factory.build()

        # run a simulation on a given system
        context = Context(root_component,
                          PhaseManager(analysis.system_template.phase_set,
                                       analysis.system_template.phase_jump_trigger_set),
                          RootCauseAnalysisManager(analysis.system_template.root_cause_analysis_trigger_set,
                                                   root_component.to_set()))

        logging.debug(f"Simulation: {str(analysis.simulation)}")
        logging.info(f"System tree: \n{str(root_component)}")
        logging.debug(f"PhaseManager: {str(context.phase_manager)}")

        simulation_results = analysis.simulation.run(context)

        execution_time = (datetime.now() - start_time).total_seconds()
        logging.info(f"Analysis execution time = {execution_time:.2f} s")

        # export results
        AggregateExporter.export(context,
                                 output_folder,
                                 analysis,
                                 simulation_results,
                                 execution_time)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class SystemModifier:

    def __init__(self,
                 parameter_name,
                 value):
        self.parameter_name = parameter_name
        self.value = value

    def __str__(self):
        return f"parameter {self.parameter_name} -> new value: {self.value}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.parameter_name == other.parameter_name \
               and self.value == other.value



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging

from availsim4core.src.context.system.system_utils import SystemUtils
from availsim4core.src.reader.xlsx.monte_carlo_reader import MonteCarloReader
from availsim4core.src.reader.xlsx.quasi_monte_carlo_reader import QuasiMonteCarloReader
from availsim4core.src.reader.xlsx.splitting_monte_carlo_reader import SplittingMonteCarloReader
from availsim4core.src.reader.xlsx.xlsx_reader import XLSXReader
from availsim4core.src.simulation.simulation import SimulationType
from availsim4core.src.statistics.critical_failure_paths import CriticalFailurePaths


class SimulationNotFoundError(Exception):
    pass

class DiagnosticNotFoundError(Exception):
    pass


class SimulationSheet:
    SHEET = "SIMULATION"
    TYPE = "SIMULATION_TYPE"


class SimulationReader(XLSXReader):

    @classmethod
    def generate_simulation(cls, simulation_file_path):
        """
        Generate a simulation based on the given XLSX file path.
        :param simulation_file_path: path to the XLSX file with the simulation configuration.
        """

        initial_simulation_dictionary = cls.read(simulation_file_path)

        simulation_parameters = initial_simulation_dictionary[SimulationSheet.SHEET][0]
        simulation_type = simulation_parameters[SimulationSheet.TYPE]
        logging.debug(f"Reading simulation = {simulation_type}")

        if SimulationType.MONTE_CARLO == simulation_type:    
            simulation = MonteCarloReader.build(simulation_parameters)
        elif SimulationType.QUASI_MONTE_CARLO == simulation_type:
            simulation = QuasiMonteCarloReader.build(simulation_parameters)
        elif SimulationType.SPLITTING_MONTE_CARLO == simulation_type:
            simulation = SplittingMonteCarloReader.build(simulation_parameters)
        else:
            message_exception = f"Simulation {simulation_type} not found"
            logging.exception(message_exception)
            raise SimulationNotFoundError(message_exception)

        for diagnostic in simulation.list_of_diagnosis:
            from availsim4core.src.exporter.aggregate_exporter import DiagnosticType
            if diagnostic not in DiagnosticType.DIAGNOSTIC_EXPORTER.keys():
                if "CRITICAL_FAILURE_PATHS" in diagnostic:
                    # the CRITICAL_FAILURE_PATHS diagnostic could have an argument between parenthesis such as:
                    # CRITICAL_FAILURE_PATHS(specific_component_name)
                    pass
                else:
                    message_exception = f"Diagnostic {diagnostic} not found in the predefined diagnosis"
                    logging.exception(message_exception)
                    raise DiagnosticNotFoundError(message_exception)

        return simulation



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_strategy import \
    SystemModifierCombinationStrategy


class SystemModifierCombinationInnerStrategy(SystemModifierCombinationStrategy):
    """
    Strategy to apply the Inner logic.

    :eg.
        Default analysis with
            - A:0, B:0, C:0
        Given:
            - Inner(A; 1)
            - Inner(B; 2,3,4)
            - Inner(C; 10)

        It results in 5 analysis :
            - A:1, B:0, C:0
            - A:0, B:2, C:0
            - A:0, B:3, C:0
            - A:0, B:4, C:0
            - A:0, B:0, C:10
    """

    @classmethod
    def execute(cls, exploration_strategy_list):
        """
        Given a list of {ExplorationStrategy} it returns a list of corresponding {SystemModifierCombination}.
        :param exploration_strategy_list the list from which the Inner strategy will be extracted and applied
        to get the corresponding combination
        :returns list of {SystemModifierCombination} corresponding to the Inner logic.
        """
        inner_exploration_strategy_list = cls.filter_exploration(exploration_strategy_list,
                                                                 Inner)
        inner_combination_list = []

        for exploration_strategy_inner in inner_exploration_strategy_list:
            for inner_value in exploration_strategy_inner.values:
                curr_system_modifier = SystemModifier(exploration_strategy_inner.parameter_name, inner_value)
                inner_combination_list.append(SystemModifierCombination([curr_system_modifier]))
        return inner_combination_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.c_event.component_c_event import ComponentCEvent
from availsim4core.src.discrete_event_simulation.event.event import Event


class OrderFailureEvent(ComponentCEvent):
    """
    class used to generate C events themselves generating B events to handle failure events
    """
    __slots__ = 'event', 'failure_mode'

    def __init__(self,
                 priority: int,
                 context: Context,
                 component: Component,
                 event: Event,
                 failure_mode: FailureMode):
        super().__init__(priority, context, component)
        self.event = event
        self.failure_mode = failure_mode

    def __eq__(self, other):
        if not isinstance(self, OrderFailureEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.event == other.event and \
               self.failure_mode == other.failure_mode

    def __hash__(self):
        return hash((type(self), self.priority, self.component, self.event, self.failure_mode))

    def __str__(self):
        return f"OrderFailureEvent:: " \
               f"priority:{self.priority} - " \
               f"component:{self.component.name} - " \
               f"failure_mode:{self.failure_mode.name}"

    def is_condition_valid(self):

        if not self.failure_mode.failure_law.is_failure_on_demand:
            # this failure mode is not a failure on demand, the next failure can be immediately ordered
            return True

        # if the code passes here, then the failure mode is of the type 'failure on demand'

        if self.context.phase_manager.current_phase not in self.failure_mode.phase_set:
            # this failure mode is a failure on demand BUT the system is not in the right phase
            return False

        if self.failure_mode.uniform_samples_for_quasi_monte_carlo:
            # the code uses Quasi Monte Carlo
            is_failing_now = self.failure_mode.failure_law.get_quantile_value(
                self.failure_mode.uniform_samples_for_quasi_monte_carlo.pop(0)
            )
            if not self.failure_mode.uniform_samples_for_quasi_monte_carlo:
                # if the list of predefined Quasi Monte Carlo samples is empty, display the information
                logging.info("The simulation is going from QMC to MC")
        else:
            # the code is not using Quasi Monte Carlo
            is_failing_now = self.failure_mode.failure_law.get_random_value()
        return is_failing_now

    def generate_b_events(self, absolute_simulation_time):
        from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.failure_event_factory import \
            FailureEventFactory

        if self.failure_mode.failure_law.is_failure_on_demand:
            time_to_fail = 0

        else:

            # -> uniform_samples_for_quasi_monte_carlo
            if self.failure_mode.uniform_samples_for_quasi_monte_carlo != []:
                time_to_fail = self.failure_mode.failure_law.get_quantile_value(
                    self.failure_mode.uniform_samples_for_quasi_monte_carlo.pop(0)
                )
                if not self.failure_mode.uniform_samples_for_quasi_monte_carlo:
                    logging.info("The simulation is going from QMC to MC")
            else:
                time_to_fail = self.failure_mode.failure_law.get_random_value()

        event = FailureEventFactory.build(
            absolute_simulation_time + time_to_fail,
            self.context,
            self.component,
            self.failure_mode
        )
        return {event}



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_Inspection(unittest.TestCase):
    """
    Ensemble of tests used to test the inspection feature
    """

    expected_result_scenario_1 = [Result(4, "A2_0_11", "NONE", "INSPECTION", "_MEAN_OCCURRENCES"),
                                  Result(4, "B4_0_8", "NONE", "INSPECTION", "_MEAN_OCCURRENCES"),
                                  Result(4, "C2_0_5", "NONE", "DEGRADED", "_MEAN_OCCURRENCES"),
                                  Result(4, "D_0_1", "NONE", "DEGRADED", "_MEAN_OCCURRENCES")
                                  ]
    scenario_1_basic_inspection_no_failure = \
        ["./availsim4core/test/E2E/input/inspection/simulation.xlsx",
         "./availsim4core/test/E2E/input/inspection/basic_inspection_no_failure_system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    expected_result_scenario_2 = [Result(4, "A2_0_11", "NONE", "INSPECTION", "_MEAN_OCCURRENCES"),
                                  Result(4, "B4_0_8", "NONE", "INSPECTION", "_MEAN_OCCURRENCES"),
                                  Result(4, "C2_0_5", "NONE", "DEGRADED", "_MEAN_OCCURRENCES"),
                                  Result(4, "D_0_1", "NONE", "DEGRADED", "_MEAN_OCCURRENCES")
                                  ]
    scenario_2_basic_inspection_before_failure = \
        ["./availsim4core/test/E2E/input/inspection/simulation.xlsx",
         "./availsim4core/test/E2E/input/inspection/basic_inspection_before_failure_system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    expected_result_scenario_3 = [Result(2, "A2_0_11", "NONE", "INSPECTION", "_MEAN_OCCURRENCES"),
                                  Result(2, "B4_0_8", "NONE", "INSPECTION", "_MEAN_OCCURRENCES"),
                                  Result(2, "C2_0_5", "NONE", "DEGRADED", "_MEAN_OCCURRENCES"),
                                  Result(2, "D_0_1", "NONE", "DEGRADED", "_MEAN_OCCURRENCES")
                                  ]

    scenario_3_basic_inspection_after_failure = \
        ["./availsim4core/test/E2E/input/inspection/simulation.xlsx",
         "./availsim4core/test/E2E/input/inspection/basic_inspection_after_failure_system.xlsx",
         None,
         1,
         expected_result_scenario_3]

    param_scenario = [
        scenario_1_basic_inspection_no_failure,
        scenario_2_basic_inspection_before_failure,
        scenario_3_basic_inspection_after_failure
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_result_list in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_result_list)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    sensitivity_file,
                    expected_number_file_result,
                    expected_result_list):
        output_folder = "./availsim4core/test/E2E/output/inspection/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)
            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import numpy


class Statistics:
    """
    Class used to gather statistics about a particular metric (number of failure, duration of failure etc.).
    Ideally histograms should be produced, for now, simpler estimators are used.
    """

    def __init__(self):
        self.min = numpy.inf
        self.max = - numpy.inf
        self.sum = 0
        self.sum_square = 0
        self.boolean = 0

        self.mean = numpy.NaN
        self.std = numpy.NaN

    def update_min(self, value):
        if self.min > value:
            self.min = value

    def update_max(self, value):
        if self.max < value:
            self.max = value

    def update_sum(self, value, weight: float = 1.0):
        self.sum = self.sum + value * weight

    def update_sum_square(self, value, weight: float = 1.0):
        # It's possible that the weight should be outside the parenthesis
        self.sum_square = self.sum_square + value ** 2 * weight

    def update_boolean(self, value, weight: float = 1.0):
        self.boolean = self.boolean + bool(value) * weight

    def update(self, value, weight: float = 1.0):
        self.update_min(value)
        self.update_max(value)
        self.update_sum(value, weight)
        self.update_sum_square(value, weight)
        self.update_boolean(value, weight)

    def evaluate_result(self):
        self.evaluate_mean()
        self.evaluate_standard_deviation()
        self.evaluate_bool()

    def evaluate_mean(self):
        self.mean = self.sum 

    def evaluate_standard_deviation(self):
        x = float(self.sum_square - (self.sum) ** 2)
        self.std = 0.0
        if x > 10e-12:
            self.std = numpy.sqrt(
                float(self.sum_square - (self.sum) ** 2))

    def evaluate_bool(self):
        self.boolean = self.boolean

    def __str__(self):
        return "MIN:{}, MAX:{}, SUM:{}, SUM_SQUARE:{}, BOOLEAN:{}".format(self.min,
                                                                          self.max,
                                                                          self.sum,
                                                                          self.sum_square,
                                                                          self.boolean)

    def __repr__(self):
        return self.__str__()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from enum import Enum


class Status(Enum):
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    DEGRADED = "DEGRADED"
    BLIND_FAILED = "BLIND_FAILED"
    BLIND_DEGRADED = "BLIND_DEGRADED"
    UNDER_REPAIR = "UNDER_REPAIR"
    INSPECTION = "INSPECTION"
    HELD = "HELD"

    def __str__(self):
        return str(self.name)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw


class Phase:
    """
    Class dealing with phases, defining each of them
    """
    __slots__ = 'name', 'law', 'next', 'is_first_phase', 'next_phase_if_failure'

    def __init__(self,
                 phase_name: str,
                 phase_law: ProbabilityLaw,
                 phase_first: bool):
        self.name = phase_name
        self.law = phase_law
        self.next = None
        self.is_first_phase = phase_first
        self.next_phase_if_failure = None

    def set_next_phase(self, next_phase):
        self.next = next_phase

    def set_next_phase_if_failure(self, next_phase_if_failure):
        self.next_phase_if_failure = next_phase_if_failure

    def __hash__(self):
        return hash((type(self), self.name))

    def __eq__(self, other):
        if not isinstance(other, Phase):
            return NotImplemented
        return self.name == other.name

    def __str__(self):
        if self.next is None or self.next_phase_if_failure is None:
            return (f"Phase name:{self.name} ~ "
                    f"law:{self.law} ~ "
                    f"first:{self.is_first_phase} ~ ")
        else:
            return (f"Phase name:{self.name} ~ "
                    f"law:{self.law} ~ "
                    f"next:{self.next.name} ~ "
                    f"first:{self.is_first_phase} ~ "
                    f"next if failure:{self.next_phase_if_failure.name}")


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import numpy
from numpy.random import normal
from scipy.special import erfinv

from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw


class NormalLaw(ProbabilityLaw):
    """
    Class used to generate values according to the Normal law
    """

    def __init__(self,
                 parameters):
        super().__init__(self.__class__.__name__,
                         parameters,
                         is_failure_on_demand=False)

    def __eq__(self, other):
        if not isinstance(other, NormalLaw):
            return NotImplemented
        return super().__eq__(other)

    def get_random_value(self):
        return normal(self.parameters[0], self.parameters[1])

    def get_quantile_value(self, quantile):
        # https://en.wikipedia.org/wiki/Normal_distribution
        return self.parameters[0] + self.parameters[1] * numpy.sqrt(2) * erfinv(2 * quantile - 1)

    def get_mean_value(self):
        return self.parameters[0]

    def get_variance_value(self):
        return self.parameters[0] ** 2



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.system.failure_factory import FailureFactory, FailureFactoryError


class test_FailureFactory(unittest.TestCase):

    def test_build_exception(self):
        with self.assertRaises(FailureFactoryError) as context:
            FailureFactory.build("impossible_value")
        self.assertTrue("Wrong type of failure" in str(context.exception))



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.timeline.record_component import RecordComponent
from availsim4core.src.timeline.record_phase import RecordPhase


class RecordGenerator:
    """
    Class handling the generation of the initial records used at the beginning of a timeline
    """

    DESCRIPTION_INIT = "Init"

    @classmethod
    def get_initial_records(cls,
                            absolute_simulation_time: float,
                            context: Context):
        """
        Function returning the initial records related to phases and components
        """
        record_list = []
        init_phase_records = cls._get_initial_phase_records(absolute_simulation_time,
                                                            context)
        init_component_records = cls._get_initial_component_records(absolute_simulation_time,
                                                                    context)
        record_list.extend(init_phase_records)
        record_list.extend(init_component_records)
        return record_list

    @classmethod
    def _get_initial_phase_records(cls,
                                   absolute_simulation_time: float,
                                   context: Context):
        return [RecordPhase(context.phase_manager.current_phase,
                            absolute_simulation_time,
                            RecordGenerator.DESCRIPTION_INIT)]

    @classmethod
    def _get_initial_component_records(cls,
                                       absolute_simulation_time: float,
                                       context: Context):
        record_list = []
        component_set = context.root_component.to_set()

        for component in component_set:
            record = RecordComponent(component,
                                     Status.RUNNING,
                                     absolute_simulation_time,
                                     context.phase_manager.current_phase,
                                     RecordGenerator.DESCRIPTION_INIT)
            record_list.append(record)
        return record_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest
from unittest.mock import patch

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.system.architecture_entry import ArchitectureEntry
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure_mode_assignments import FailureModeAssignments
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.context.system.probability_law.exponential_law import ExponentialLaw
from availsim4core.src.context.system.system_template import SystemTemplate
from availsim4core.src.reader.xlsx.sensitivity_analysis_reader import SensitivityAnalysisReader
from availsim4core.src.sensitivity_analysis.sensitivity_analysis import SensitivityAnalysis
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.simulation.monte_carlo import MonteCarlo


class test_SensitivityAnalysis(unittest.TestCase):

    def test_generate_analysis_list_1(self):
        SENSITIVITY_SEED_1 = 10
        SENSITIVITY_SEED_2 = 100

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   [],
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager, 
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 1 -> SENSITIVITY_SEED_1
        analysis_simulation_1 = MonteCarlo(1000, 1000, 2., 120, SENSITIVITY_SEED_1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   [],
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_1 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_1 = Analysis(0, analysis_system_template_1, analysis_simulation_1)

        # Analysis 2 -> SENSITIVITY_SEED_2
        analysis_simulation_2 = MonteCarlo(1000, 1000, 2., 120, SENSITIVITY_SEED_2, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   [],
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_2 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_2 = Analysis(1, analysis_system_template_2, analysis_simulation_2)

        expected_result = [analysis_1, analysis_2]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'SEED',
                    "VALUES": f'[{SENSITIVITY_SEED_1}, {SENSITIVITY_SEED_2}]',
                    "EXPLORATION_STRATEGY": "INNER"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_2(self):
        SENSITIVITY_FAILURE_1 = [8.]
        SENSITIVITY_FAILURE_2 = [9.]

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 3 -> SENSITIVITY_FAILURE_1
        analysis_simulation_3 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_1),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_3 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager, 
                                                    set(),
                                                    set(),
                                                    None)
        analysis_3 = Analysis(0, analysis_system_template_3, analysis_simulation_3)

        # Analysis 4 -> SENSITIVITY_FAILURE_2
        analysis_simulation_4 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_2),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_4 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_4 = Analysis(1, analysis_system_template_4, analysis_simulation_4)

        expected_result = [analysis_3, analysis_4]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": f'[{SENSITIVITY_FAILURE_1}, {SENSITIVITY_FAILURE_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_3(self):
        SENSITIVITY_COMPONENT_NUMBER_1 = 4
        SENSITIVITY_COMPONENT_NUMBER_2 = 5

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 3 -> SENSITIVITY_COMPONENT_NUMBER_1
        analysis_simulation_3 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_3 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_3 = Analysis(0, analysis_system_template_3, analysis_simulation_3)

        # Analysis 5 -> SENSITIVITY_COMPONENT_NUMBER_2
        analysis_simulation_5 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_2, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_5 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_5 = Analysis(1, analysis_system_template_5, analysis_simulation_5)

        expected_result = [analysis_3, analysis_5]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": f'[{SENSITIVITY_COMPONENT_NUMBER_1}, {SENSITIVITY_COMPONENT_NUMBER_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"},
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_4(self):

        def get_system_template_and_simulation(SEED,
                                               FAILURE,
                                               COMPONENT_NUMBER,
                                               CHILDREN_LOGIC):

            initial_simulation = MonteCarlo(1000, 1000, 2., 120, SEED, ['SUMMARY'], 1)

            architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], CHILDREN_LOGIC, [], []),
                                       ArchitectureEntry("DUMBER", "BASIC", COMPONENT_NUMBER, [], '', [], [])]
            failure_mode = FailureMode("DUMBERFAILURE",
                                       ExponentialLaw(FAILURE),
                                       DeterministicLaw([1000]),
                                       Failure(FailureType.BLIND),
                                       "[]",
                                       None,
                                       [],
                                       None,
                                       'NEVER',
                                       []
                                       )

            failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
            mru_list = []
            inspection_list = []
            phase_manager = None

            initial_system_template = SystemTemplate(architecture_entry_list,
                                                     failure_mode_assignment_list,
                                                     [failure_mode],
                                                     mru_list,
                                                     inspection_list,
                                                     phase_manager,
                                                     set(),
                                                     set(),
                                                     None)

            return initial_system_template, initial_simulation

        def get_new_analysis(id_analysis,
                             SEED,
                             FAILURE,
                             COMPONENT_NUMBER,
                             CHILDREN_LOGIC):

            system_template, simulation = get_system_template_and_simulation(SEED,
                                                                             FAILURE,
                                                                             COMPONENT_NUMBER,
                                                                             CHILDREN_LOGIC)

            return Analysis(id_analysis, system_template, simulation)

        SEED = 1
        SENSITIVITY_SEED_1 = 10
        SENSITIVITY_SEED_2 = 100

        FAILURE = [7.]
        SENSITIVITY_FAILURE_1 = [8.]
        SENSITIVITY_FAILURE_2 = [9.]

        COMPONENT_NUMBER = 3
        SENSITIVITY_COMPONENT_NUMBER_1 = 4
        SENSITIVITY_COMPONENT_NUMBER_2 = 5

        CHILDREN_LOGIC = 'AND'
        SENSITIVITY_CHILDREN_LOGIC_1 = '1RC'
        SENSITIVITY_CHILDREN_LOGIC_2 = '1TF'

        initial_system_template, initial_simulation = get_system_template_and_simulation(SEED,
                                                                                         FAILURE,
                                                                                         COMPONENT_NUMBER,
                                                                                         CHILDREN_LOGIC)

        expected_result = [

            get_new_analysis(0,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_1,
                             SENSITIVITY_COMPONENT_NUMBER_1,
                             CHILDREN_LOGIC),
            get_new_analysis(1,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_1,
                             SENSITIVITY_COMPONENT_NUMBER_1,
                             CHILDREN_LOGIC),
            get_new_analysis(2,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_2,
                             SENSITIVITY_COMPONENT_NUMBER_1,
                             CHILDREN_LOGIC),
            get_new_analysis(3,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_2,
                             SENSITIVITY_COMPONENT_NUMBER_1,
                             CHILDREN_LOGIC),

            get_new_analysis(4,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_1,
                             SENSITIVITY_COMPONENT_NUMBER_2,
                             CHILDREN_LOGIC),
            get_new_analysis(5,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_1,
                             SENSITIVITY_COMPONENT_NUMBER_2,
                             CHILDREN_LOGIC),
            get_new_analysis(6,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_2,
                             SENSITIVITY_COMPONENT_NUMBER_2,
                             CHILDREN_LOGIC),
            get_new_analysis(7,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_2,
                             SENSITIVITY_COMPONENT_NUMBER_2,
                             CHILDREN_LOGIC),


            get_new_analysis(8,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_1,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_1),
            get_new_analysis(9,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_1,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_1),
            get_new_analysis(10,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_2,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_1),
            get_new_analysis(11,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_2,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_1),

            get_new_analysis(12,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_1,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_2),
            get_new_analysis(13,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_1,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_2),
            get_new_analysis(14,
                             SENSITIVITY_SEED_1,
                             SENSITIVITY_FAILURE_2,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_2),
            get_new_analysis(15,
                             SENSITIVITY_SEED_2,
                             SENSITIVITY_FAILURE_2,
                             COMPONENT_NUMBER,
                             SENSITIVITY_CHILDREN_LOGIC_2)
        ]


        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": f'[{SENSITIVITY_COMPONENT_NUMBER_1}, {SENSITIVITY_COMPONENT_NUMBER_2}]',
                    "EXPLORATION_STRATEGY": "INNER"},
                1: {"PARAMETER_NAME": 'CHILDREN_LOGIC/DUMB',
                    "VALUES": f'[{SENSITIVITY_CHILDREN_LOGIC_1}, {SENSITIVITY_CHILDREN_LOGIC_2}]',
                    "EXPLORATION_STRATEGY": "INNER"},
                2: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": f'[{SENSITIVITY_FAILURE_1}, {SENSITIVITY_FAILURE_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"},
                3: {"PARAMETER_NAME": 'SEED',
                    "VALUES": f'[{SENSITIVITY_SEED_1}, {SENSITIVITY_SEED_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_5(self):
        CHILDREN_LOGIC_1 = "3OO4"
        CHILDREN_LOGIC_2 = "2OO4"

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 4, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 1 -> CHILDREN_LOGIC_1
        analysis_simulation_1 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], CHILDREN_LOGIC_1, [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 4, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_1 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_1 = Analysis(0, analysis_system_template_1, analysis_simulation_1)

        # Analysis 2 -> CHILDREN_LOGIC_2
        analysis_simulation_2 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], CHILDREN_LOGIC_2, [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 4, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_2 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_2 = Analysis(1, analysis_system_template_2, analysis_simulation_2)

        expected_result = [analysis_1, analysis_2]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'CHILDREN_LOGIC/DUMB',
                    "VALUES": f'[{CHILDREN_LOGIC_1}, {CHILDREN_LOGIC_2}]',
                    "EXPLORATION_STRATEGY": "INNER"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_6(self):
        CHILDREN_LOGIC_1 = "3OO4"
        CHILDREN_LOGIC_2 = "2OO4"

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 4, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 1 -> CHILDREN_LOGIC_1
        analysis_simulation_1 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], CHILDREN_LOGIC_1, [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 4, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_1 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager, 
                                                    set(),
                                                    set(),
                                                    None)
        analysis_1 = Analysis(0, analysis_system_template_1, analysis_simulation_1)

        # Analysis 2 -> CHILDREN_LOGIC_2
        analysis_simulation_2 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], CHILDREN_LOGIC_2, [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 4, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_2 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_2 = Analysis(1, analysis_system_template_2, analysis_simulation_2)

        expected_result = [analysis_1, analysis_2]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'CHILDREN_LOGIC/DUMB',
                    "VALUES": f'[{CHILDREN_LOGIC_1}, {CHILDREN_LOGIC_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_7(self):
        # only outer strategy

        SENSITIVITY_FAILURE_1 = [8.]
        SENSITIVITY_FAILURE_2 = [9.]
        SENSITIVITY_COMPONENT_NUMBER_1 = 4
        SENSITIVITY_COMPONENT_NUMBER_2 = 5

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 3 -> SENSITIVITY_COMPONENT_NUMBER_1 / SENSITIVITY_FAILURE_1
        analysis_simulation_3 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_1),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_3 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_3 = Analysis(0, analysis_system_template_3, analysis_simulation_3)

        # Analysis 4 -> SENSITIVITY_COMPONENT_NUMBER_1 / SENSITIVITY_FAILURE_2
        analysis_simulation_4 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_2),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_4 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_4 = Analysis(1, analysis_system_template_4, analysis_simulation_4)

        # Analysis 5 -> SENSITIVITY_COMPONENT_NUMBER_2 / SENSITIVITY_FAILURE_1
        analysis_simulation_5 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_2, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_1),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_5 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_5 = Analysis(2, analysis_system_template_5, analysis_simulation_5)

        # Analysis 6 -> SENSITIVITY_COMPONENT_NUMBER_2 / SENSITIVITY_FAILURE_2
        analysis_simulation_6 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_2, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_2),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_6 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_6 = Analysis(3, analysis_system_template_6, analysis_simulation_6)

        expected_result = [analysis_3, analysis_4, analysis_5, analysis_6]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": f'[{SENSITIVITY_COMPONENT_NUMBER_1}, {SENSITIVITY_COMPONENT_NUMBER_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"},
                1: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": f'[{SENSITIVITY_FAILURE_1}, {SENSITIVITY_FAILURE_2}]',
                    "EXPLORATION_STRATEGY": "OUTER"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_8(self):
        # only zip strategy
        # two elements in the sensitivity analysis

        SENSITIVITY_FAILURE_1 = [8.]
        SENSITIVITY_FAILURE_2 = [9.]
        SENSITIVITY_COMPONENT_NUMBER_1 = 4
        SENSITIVITY_COMPONENT_NUMBER_2 = 5

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 0 -> SENSITIVITY_COMPONENT_NUMBER_1 / SENSITIVITY_FAILURE_1
        analysis_simulation_0 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_1),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_0 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_0 = Analysis(0, analysis_system_template_0, analysis_simulation_0)

        # Analysis 1 -> SENSITIVITY_COMPONENT_NUMBER_2 / SENSITIVITY_FAILURE_2
        analysis_simulation_1 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_2, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_2),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_1 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_1 = Analysis(1, analysis_system_template_1, analysis_simulation_1)

        expected_result = [analysis_0, analysis_1]
        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": f'[{SENSITIVITY_COMPONENT_NUMBER_1}, {SENSITIVITY_COMPONENT_NUMBER_2}]',
                    "EXPLORATION_STRATEGY": "ZIP"},
                1: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": f'[{SENSITIVITY_FAILURE_1}, {SENSITIVITY_FAILURE_2}]',
                    "EXPLORATION_STRATEGY": "ZIP"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis_list_9(self):
        # only zip strategy
        # one element in the sensitivity analysis

        SENSITIVITY_FAILURE_1 = [8.]
        SENSITIVITY_COMPONENT_NUMBER_1 = 4

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis 0 -> SENSITIVITY_COMPONENT_NUMBER_1 / SENSITIVITY_FAILURE_1
        analysis_simulation_0 = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_1),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_0 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        analysis_0 = Analysis(0, analysis_system_template_0, analysis_simulation_0)

        expected_result = [analysis_0]

        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": f'[{SENSITIVITY_COMPONENT_NUMBER_1}]',
                    "EXPLORATION_STRATEGY": "ZIP"},
                1: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": f'[{SENSITIVITY_FAILURE_1}]',
                    "EXPLORATION_STRATEGY": "ZIP"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                       initial_system_template)
            result = sensitivity_analysis.generate_analysis_list("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_analysis__system(self):
        SENSITIVITY_FAILURE_1 = [8.]
        SENSITIVITY_COMPONENT_NUMBER_1 = 4

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Analysis -> SENSITIVITY_FAILURE_1 / SENSITIVITY_COMPONENT_NUMBER_1
        analysis_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", SENSITIVITY_COMPONENT_NUMBER_1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw(SENSITIVITY_FAILURE_1),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        analysis_system_template_2 = SystemTemplate(architecture_entry_list,
                                                    failure_mode_assignment_list,
                                                    [failure_mode],
                                                    mru_list,
                                                    inspection_list,
                                                    phase_manager,
                                                    set(),
                                                    set(),
                                                    None)
        expected_analysis = Analysis(1, analysis_system_template_2, analysis_simulation)

        system_modifier_failure_1 = SystemModifier("FAILURE_PARAMETERS/DUMBERFAILURE", SENSITIVITY_FAILURE_1)
        system_modifier_component_number_1 = SystemModifier("COMPONENT_NUMBER/DUMBER", SENSITIVITY_COMPONENT_NUMBER_1)

        sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                   initial_system_template)

        result_analysis = sensitivity_analysis._generate_analysis(1,
                                                                  SystemModifierCombination(
                                                                      [system_modifier_failure_1,
                                                                       system_modifier_component_number_1]))

        self.assertEqual(expected_analysis, result_analysis)

    def test_apply_combination_modifier_on_system(self):
        # Initial simulation / Systems
        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 1, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   [],
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        initial_system_template = SystemTemplate(architecture_entry_list,
                                                 failure_mode_assignment_list,
                                                 [failure_mode],
                                                 mru_list,
                                                 inspection_list,
                                                 phase_manager,
                                                 set(),
                                                 set(),
                                                 None)

        # Expected
        architecture_entry_list = [ArchitectureEntry("DUMB", "COMPOUND", 1, ['DUMBER'], "AND", [], []),
                                   ArchitectureEntry("DUMBER", "BASIC", 100, [], '', [], [])]
        failure_mode = FailureMode("DUMBERFAILURE",
                                   ExponentialLaw([1.]),
                                   DeterministicLaw([1000]),
                                   Failure(FailureType.BLIND),
                                   [],
                                   None,
                                   [],
                                   None,
                                   'NEVER',
                                   []
                                   )

        failure_mode_assignment_list = [FailureModeAssignments("DUMBER", failure_mode)]
        mru_list = []
        inspection_list = []
        phase_manager = None

        expected_result = SystemTemplate(architecture_entry_list,
                                         failure_mode_assignment_list,
                                         [failure_mode],
                                         mru_list,
                                         inspection_list,
                                         phase_manager,
                                         set(),
                                         set(),
                                         None)

        system_modifier_failure_1 = SystemModifier("COMPONENT_NUMBER/DUMBER", 100)
        system_modifier_combination = SystemModifierCombination([system_modifier_failure_1])

        sensitivity_analysis = SensitivityAnalysis(None,
                                                   initial_system_template)
        result = sensitivity_analysis._apply_combination_modifier_on_system(system_modifier_combination)

        self.assertEqual(expected_result, result)

    def test_apply_combination_modifier_on_simulation(self):
        SENSITIVITY_SEED = 100

        expected_result = MonteCarlo(1000, 1000, 2., 120, SENSITIVITY_SEED, ['SUMMARY'], 1)

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        system_modifier_failure_1 = SystemModifier("SEED", SENSITIVITY_SEED)
        system_modifier_combination = SystemModifierCombination([system_modifier_failure_1])

        sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                   None)
        result = sensitivity_analysis._apply_combination_modifier_on_simulation(system_modifier_combination)

        self.assertEqual(expected_result, result)

    def test_generate_analysis__simulation(self):
        SENSITIVITY_SEED_1 = 100

        # Initial simulation / Systems
        initial_simulation = MonteCarlo(1000, 1000, 2., 120, 1, ['SUMMARY'], 1)

        initial_system_template = SystemTemplate([], [], [], [], [], None, set(),set(),None)

        # Analysis -> SENSITIVITY_SEED_1
        analysis_simulation = MonteCarlo(1000, 1000, 2., 120, SENSITIVITY_SEED_1, ['SUMMARY'], 1)

        analysis_1 = Analysis(0, initial_system_template, analysis_simulation)

        system_modifier_seed = SystemModifier("SEED", SENSITIVITY_SEED_1)

        sensitivity_analysis = SensitivityAnalysis(initial_simulation,
                                                   initial_system_template)

        result_analysis_1 = sensitivity_analysis._generate_analysis(0,
                                                                    SystemModifierCombination([system_modifier_seed]))

        self.assertEqual(analysis_1, result_analysis_1)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterLastTimeline(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        # convert the last timeline for the conversion into a pandas dataframe
        timeline_export = []
        for record in results.last_simulation_timeline:
            timeline_export.append(
                {"timestamp": record.timestamp,
                 "phase": record.phase.name,
                 "component": record.get_result_record_entry(),
                 "description": record.description}
            )
        timeline_export_dataframe = pandas.DataFrame(timeline_export)
        timeline_export_dataframe.to_excel(writer, sheet_name='RESULTS_LAST_TIMELINE', index=False)

        writer.save()


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context


class Event:
    """
    Global Interface to define an event (b_event or c_event)
    event are processed in the Discrete event simulation algorithm.
    Events are immutable objects.
    """
    __slots__ = 'context'

    def __init__(self,
                 context: Context):
        self.context = context


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List

from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.context.system.probability_law.exponential_law import ExponentialLaw
from availsim4core.src.context.system.probability_law.normal_law import NormalLaw
from availsim4core.src.context.system.probability_law.binomial_law import BinomialLaw
from availsim4core.src.context.system.probability_law.weibull_law import WeibullLaw


class ProbabilityLawFactoryError(Exception):
    pass


class ProbabilityLawFactory:

    @staticmethod
    def build(distribution_str,
              parameters: List[float]):

        if distribution_str in ["EXP", "EXPONENTIAL", "EXPONENTIALLAW", "EXPONENTIAL_LAW"]:
            return ExponentialLaw(parameters)
        elif distribution_str in ["NORMAL", "NORMALLAW", "NORMAL_LAW"]:
            return NormalLaw(parameters)
        elif distribution_str in ["FIX", "DETERMINISTIC", "DETERMINISTIC_LAW", "DETERMINISTICLAW"]:
            return DeterministicLaw(parameters)
        elif distribution_str in ["POFOD", "FOD","BINOMIAL", "BINOMIALLAW", "BINOMIAL_LAW"]:
            return BinomialLaw(parameters)
        elif distribution_str in ["WEIBULL", "WEIBULLLAW", "WEIBULL_LAW"]:
            return WeibullLaw(parameters)
        else:
            message_exception = f"wrong type of distribution function: {distribution_str}"
            import logging
            logging.exception(message_exception)
            raise ProbabilityLawFactoryError(message_exception)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import Set

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.phase_jump_trigger import PhaseJumpTrigger
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority
from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import NextPhaseEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_jump_phase_event import \
    OrderJumpPhaseEvent


class JumpPhaseEvent(NextPhaseEvent):
    """
    Class handling the change of phases
    """
    __slots__ = 'phase_jump_trigger'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 phase_jump_trigger: PhaseJumpTrigger):
        super().__init__(absolute_occurrence_time,context,
                         BEventPriority.JUMP_PHASE_EVENT)
        self.phase_jump_trigger = phase_jump_trigger

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.phase_jump_trigger))

    def __eq__(self, other):
        if not isinstance(other, JumpPhaseEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.phase_jump_trigger == other.phase_jump_trigger

    def __str__(self):
        return (f"type ::{type(self)} "
                f"due to ::{self.phase_jump_trigger} "
                f"at t ::{self.absolute_occurrence_time} \n")

    def execute(self):

        records = [self.context.phase_manager.go_to_phase(
                self.absolute_occurrence_time,
                self.phase_jump_trigger.to_phase,
                f"Jump triggered by {self.phase_jump_trigger.component_name} "\
                f" with status {self.phase_jump_trigger.component_status} "\
                f" from phase {self.context.phase_manager.current_phase.name} "\
                f" to phase {self.phase_jump_trigger.to_phase.name}")]

        records.extend(self.context.add_record_of_each_component(
            self.absolute_occurrence_time,
            self.context.phase_manager.UPDATE_FOLLOWING_A_PHASE_CHANGE))

        return records

    def generate_c_event(self, **kwargs) -> Set:

        c_event_set = NextPhaseEvent.generate_c_event(self,**kwargs)

        #TODO regenerate only one c_event or all?

        c_event_set.add(
            OrderJumpPhaseEvent(
                priority=CEventPriority.ORDER_JUMP_PHASE_EVENT,
                context=self.context,
                phase_jump_trigger=self.phase_jump_trigger
            )
        )

        return c_event_set



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.timeline.record_component import RecordComponent


class test_Component(unittest.TestCase):

    def test__eq__(self):
        compound_1 = Compound(1, "test_compound_1", 0, [], ChildrenLogic(), [])
        compound_2 = Compound(1, "test_compound_1", 0, [], ChildrenLogic(), [])

        self.assertEqual(compound_1, compound_2)

    def test_update_status__single_new_status(self):

        context = Context(None, None, None)

        component = Basic(1, "test_compound_1", 1, [], [], [], [])

        new_status = Status.FAILED
        timestamp = 916
        phase = PhaseManager.NO_PHASE
        description = "test"

        result = component._update_status(new_status, timestamp, phase, description, context)
        expected_result = [RecordComponent(component, new_status, timestamp, phase, description)]
        self.assertEqual(result, expected_result)
        self.assertEqual(component.status, new_status)

    def test_update_status__single_same_status(self):

        context = Context(None, None, None)

        component = Basic(1, "test_compound_1", 1, [], [], [], [])

        new_status = Status.RUNNING
        timestamp = 916
        phase = PhaseManager.NO_PHASE
        description = "test"

        result = component._update_status(new_status, timestamp, phase, description, context)
        # new_status of the component is the same as the previous one, no record.
        expected_result_not_in = RecordComponent(component, new_status, timestamp, phase, description)
        self.assertEqual(component.status, new_status)
        self.assertNotIn(expected_result_not_in, result)

    def test_update_status__multiple_status_sorted_by_timestamp(self):

        context = Context(None, None, None)

        component = Basic(1, "test_compound_1", 1, [], [], [], [])

        new_status_1 = Status.FAILED
        timestamp_1 = 996
        phase = PhaseManager.NO_PHASE
        description_1 = "test_failed"
        record1 = component.update_status(new_status_1, timestamp_1, phase, description_1, context)

        new_status_2 = Status.UNDER_REPAIR
        timestamp_2 = 998
        phase = PhaseManager.NO_PHASE
        description_2 = "test_repair"
        record2 = component._update_status(new_status_2, timestamp_2, phase, description_2, context)

        new_status_3 = Status.RUNNING
        timestamp_3 = 1098
        phase = PhaseManager.NO_PHASE
        description_3 = "test_running"
        record3 = component._update_status(new_status_3, timestamp_3, phase, description_3, context)

        self.assertEqual(component.status, new_status_3)
        expected_record_1 = RecordComponent(component, new_status_1, timestamp_1, phase, description_1)
        expected_record_2 = RecordComponent(component, new_status_2, timestamp_2, phase, description_2)
        expected_record_3 = RecordComponent(component, new_status_3, timestamp_3, phase, description_3)
        self.assertIn(expected_record_1, record1)
        self.assertIn(expected_record_2, record2)
        self.assertIn(expected_record_3, record3)

    def test_update_status__multiple_parents__shared_child(self):
        """
        one child is shared by two parents with and logic
        the status of the child is updated (set and propagate) to FAILED
        the status of the parents has to change too
        """

        context = Context(None, None, None)

        parent0 = Compound(1, "parent", 0, [], And(), [])
        parent1 = Compound(2, "parent", 1, [], And(), [])
        parent2 = Compound(3, "parent", 2, [], And(), [])

        shared_child = Basic(4, "shared_child", 1, [parent0, parent1, parent2], [], [], [])

        parent0.add_children_list([shared_child])
        parent1.add_children_list([shared_child])
        parent2.add_children_list([shared_child])

        shared_child._update_status(Status.FAILED, PhaseManager.NO_PHASE, 3.14, "anything", context)

        self.assertEqual(parent0.status, Status.FAILED)
        self.assertEqual(parent1.status, Status.FAILED)
        self.assertEqual(parent2.status, Status.FAILED)

    def test_update_status__multiple_parents__shared_child__extra_child(self):
        """
        one child is shared by two parents with and logic, and one extra child has only one parent (parent1)
        the status of the extra child is updated (set and propagate) to FAILED
        only the status of the parent1 has to change
        """

        context = Context(None, None, None)

        parent0 = Compound(1, "parent", 0, [], And(), [])
        parent1 = Compound(2, "parent", 1, [], And(), [])

        shared_child = Basic(3, "shared_child", 1, [parent0, parent1], [], [], [])
        child_of_parent1 = Basic(4, "child_of_parent1", 1, [parent1], [], [], [])

        parent0.add_children_list([shared_child])
        parent1.add_children_list([shared_child, child_of_parent1])

        child_of_parent1._update_status(Status.FAILED, PhaseManager.NO_PHASE, 3.14, "anything", context)

        self.assertEqual(parent0.status, Status.RUNNING)
        self.assertEqual(parent1.status, Status.FAILED)

    def test_update_status__multiple_parents__shared_children__extra_child(self):
        """
        one child is shared by two parents with and logic, and one extra child has only one parent (parent1)
        the status of the extra child is updated (set and propagate) to FAILED
        only the status of the parent1 has to change
        """

        context = Context(None, None, None)

        parent0 = Compound(1, "parent", 0, [], And(), [])
        parent1 = Compound(2, "parent", 1, [], And(), [])
        parent2 = Compound(3, "parent", 2, [], And(), [])

        shared_child0 = Basic(4, "shared_child", 0, [parent0, parent1, parent2], [], [], [])
        shared_child1 = Basic(5, "shared_child", 1, [parent0, parent1, parent2], [], [], [])
        child_of_parent1 = Basic(6, "child_of_parent1", 0, [parent1], [], [], [])

        parent0.add_children_list([shared_child0, shared_child1])
        parent1.add_children_list([shared_child0, shared_child1, child_of_parent1])
        parent2.add_children_list([shared_child0, shared_child1])

        child_of_parent1._update_status(Status.FAILED, PhaseManager.NO_PHASE, 3.14, "anything", context)
        self.assertEqual(parent0.status, Status.RUNNING)
        self.assertEqual(parent1.status, Status.FAILED)
        self.assertEqual(parent2.status, Status.RUNNING)

        shared_child0._update_status(Status.FAILED, PhaseManager.NO_PHASE, 3.14, "anything", context)
        self.assertEqual(parent0.status, Status.FAILED)
        self.assertEqual(parent1.status, Status.FAILED)
        self.assertEqual(parent2.status, Status.FAILED)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.failure import FailureType, Failure
from availsim4core.src.context.system.failure_mode import FailureMode, FailureModeError
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw


class test_FailureFactory(unittest.TestCase):

    def test_failure_mode_exception(self):
        """
        This test check that AvailSim4 will detect the combination of type_of_failure = "BLIND" and phase_change_timing = "AFTER_REPAIR"
        """
        with self.assertRaises(FailureModeError) as context:

            NO_PHASE = Phase("NONE", ProbabilityLaw("", 0, False), False)
            NO_PHASE.set_next_phase(NO_PHASE)
            NO_PHASE.set_next_phase_if_failure(NO_PHASE)

            expected_inspection = Inspection("inspection_test", 1, 2)

            FailureMode("TEST_FAILURE",
                        ProbabilityLaw("", 0, False),
                        ProbabilityLaw("", 0, False),
                        Failure(FailureType.BLIND),
                        "[]",
                        expected_inspection,
                        [],
                        NO_PHASE,
                        'AFTER_REPAIR',
                        []
                    )

        self.assertTrue("Wrong combination of type_of_failure" in str(context.exception))
